#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.dataclasses.node_info import NodeInfo


class CurrentNodeInfo(NodeInfo):
    def __init__(self, node_id, node_name, base_port, bind_ip, ima_mainnet=None,
                 ima_mp_schain=None, ima_mp_mainnet=None, wallets=None, rotate_after_block=64,
                 schain_log_level='trace', schain_log_level_config='trace'):
        self.bind_ip = bind_ip
        self.schain_log_level = schain_log_level
        self.schain_log_level_config = schain_log_level_config

        self.ima_mainnet = ima_mainnet
        self.ima_mp_schain = ima_mp_schain
        self.ima_mp_mainnet = ima_mp_mainnet
        self.wallets = wallets

        self.rotate_after_block = rotate_after_block
        super().__init__(node_id, node_name, base_port)

    def to_config(self):
        config = super().to_config()
        config['bindIP'] = self.bind_ip
        config['logLevel'] = self.schain_log_level
        config['logLevelConfig'] = self.schain_log_level_config

        config['imaMainNet'] = self.ima_mainnet
        config['imaMessageProxySChain'] = self.ima_mp_schain
        config['imaMessageProxyMainNet'] = self.ima_mp_mainnet
        config['wallets'] = self.wallets

        config['rotateAfterBlock'] = self.rotate_after_block
        return config



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE Contract manager class """

from Crypto.Hash import keccak

from skale.contracts import BaseContract
from skale.utils.helper import add_0x_prefix


class ContractManager(BaseContract):
    def get_contract_address(self, name):
        contract_hash = add_0x_prefix(self.get_contract_hash_by_name(name))
        return self.contract.functions.contracts(contract_hash).call()

    def get_contract_hash_by_name(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.hexdigest()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE helper utilities """

import ipaddress
import json
import logging
import random
import socket
import string
import sys
from logging import Formatter, StreamHandler
from random import randint


logger = logging.getLogger(__name__)


def format_fields(fields):
    """
        Transform array to object with passed fields
        Usage:
        @format(['field_name1', 'field_name2'])
        def my_method()
            return [0, 'Test']

        => {'field_name1': 0, 'field_name2': 'Test'}
    """

    def real_decorator(function):
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)

            if result is None:
                return None

            obj = {}
            for i, field in enumerate(fields):
                obj[field] = result[i]
            return obj

        return wrapper

    return real_decorator


def ip_from_bytes(bytes):
    return socket.inet_ntoa(bytes)


def ip_to_bytes(ip):  # pragma: no cover
    return socket.inet_aton(ip)


def is_valid_ipv4_address(address):
    try:
        ipaddress.IPv4Address(address)
    except ValueError:
        return False
    return True


def get_abi(abi_filepath=None):
    with open(abi_filepath) as data_file:
        return json.load(data_file)


def generate_nonce():  # pragma: no cover
    return randint(0, 65534)


def random_string(size=6, chars=string.ascii_lowercase):  # pragma: no cover
    return ''.join(random.choice(chars) for x in range(size))


def generate_random_ip():  # pragma: no cover
    return '.'.join('%s' % random.randint(0, 255) for i in range(4))


def generate_random_name(len=8):  # pragma: no cover
    return ''.join(
        random.choices(string.ascii_uppercase + string.digits, k=len))


def generate_random_port():  # pragma: no cover
    return random.randint(0, 60000)


def generate_custom_config(ip, ws_port):
    if not ip or not ws_port:
        raise ValueError(
            f'For custom init you should provide ip and ws_port: {ip}, {ws_port}'
        )
    return {
        'ip': ip,
        'ws_port': ws_port,
    }


def add_0x_prefix(bytes_string):  # pragma: no cover
    return '0x' + bytes_string


def init_default_logger():  # pragma: no cover
    handlers = []
    formatter = Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    stream_handler = StreamHandler(sys.stderr)
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    handlers.append(stream_handler)

    logging.basicConfig(level=logging.DEBUG, handlers=handlers)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from skale.dataclasses.tx_res import TxRes
from skale.utils.web3_utils import get_eth_nonce

logger = logging.getLogger(__name__)


def make_call(method, opts):
    data = method.call(opts)
    return TxRes(data=data)


def build_tx_dict(method, gas_limit, gas_price=None, nonce=None):
    tx_dict_fields = {
        'gas': gas_limit,
        'nonce': nonce
    }
    if gas_price is not None:
        tx_dict_fields.update({'gasPrice': gas_price})

    return method.buildTransaction(tx_dict_fields)


def post_transaction(wallet, method, gas_limit, gas_price=None, nonce=None) -> TxRes:
    tx_dict = build_tx_dict(method, gas_limit, gas_price, nonce)
    tx_hash = wallet.sign_and_send(tx_dict)
    return TxRes(tx_hash=tx_hash)


def sign_and_send(web3, method, gas_amount, wallet):
    nonce = get_eth_nonce(web3, wallet.address)
    tx_dict = build_tx_dict(method, gas_amount, nonce)
    signed_tx = wallet.sign(tx_dict)
    tx = web3.eth.sendRawTransaction(signed_tx.rawTransaction)
    return tx


def send_eth(web3, account, amount, wallet):
    eth_nonce = get_eth_nonce(web3, wallet.address)
    logger.info(f'Transaction nonce {eth_nonce}')
    txn = {
        'to': account,
        'from': wallet.address,
        'value': amount,
        'gasPrice': web3.eth.gasPrice,
        'gas': 22000,
        'nonce': eth_nonce
    }
    signed_txn = wallet.sign(txn)
    tx = web3.eth.sendRawTransaction(signed_txn.rawTransaction)

    logger.info(
        f'ETH transfer {wallet.address} => {account}, {amount} wei,'
        f'tx: {web3.toHex(tx)}'
    )
    return tx



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.dataclasses import CurrentNodeInfo, SchainNodeInfo
from skale.schain_config.ports_allocation import get_schain_base_port_on_node
from skale.utils.helper import ip_from_bytes
from skale.utils.web3_utils import public_key_to_address
from skale.schain_config.base_config import update_base_config


def generate_schain_info(schain, schain_nodes):
    return {
        'schainID': 1,  # todo: remove this later (should be removed from the skaled first)
        'schainName': schain['name'],
        'schainOwner': schain['owner'],
        'nodes': schain_nodes
    }


def get_nodes_for_schain(skale, name):
    nodes = []
    ids = skale.schains_data.get_node_ids_for_schain(name)
    for id_ in ids:
        node = skale.nodes_data.get(id_)
        node['id'] = id_
        nodes.append(node)
    return nodes


def get_nodes_for_schain_config(skale, name):
    nodes_info = []
    nodes = get_nodes_for_schain(skale, name)
    for i, node in enumerate(nodes, 1):
        pk = node['publicKey'].hex()

        schains_on_node = skale.schains_data.get_schains_for_node(node['id'])
        base_port = get_schain_base_port_on_node(schains_on_node, name, node['port'])

        node_info = SchainNodeInfo(
            node_name=node['name'],
            node_id=node['id'],
            base_port=base_port,

            schain_index=i,
            ip=ip_from_bytes(node['ip']),
            public_key=pk,
            public_ip=ip_from_bytes(node['publicIP']),
            owner=public_key_to_address(pk)
        ).to_config()
        nodes_info.append(node_info)
    return nodes_info


def generate_schain_config(base_config, node_info, schain_info):
    return {
        **base_config,
        'skaleConfig': {
            'nodeInfo': node_info,
            'sChain': schain_info,
        }
    }


def generate_skale_schain_config(skale, schain_name, node_id, base_config=None, ima_mainnet=None,
                                 ima_mp_schain=None, ima_mp_mainnet=None, wallets=None,
                                 rotate_after_block=64, schain_log_level='trace',
                                 schain_log_level_config='trace'):
    node = skale.nodes_data.get(node_id)
    schain = skale.schains_data.get_by_name(schain_name)

    schains_on_node = skale.schains_data.get_schains_for_node(node_id)
    schain_base_port_on_node = get_schain_base_port_on_node(schains_on_node, schain_name,
                                                            node['port'])
    schain_nodes = get_nodes_for_schain_config(skale, schain_name)

    node_info = CurrentNodeInfo(
        node_name=node['name'],
        node_id=node_id,
        base_port=schain_base_port_on_node,
        bind_ip=ip_from_bytes(node['ip']),
        ima_mainnet=ima_mainnet,
        ima_mp_schain=ima_mp_schain,
        ima_mp_mainnet=ima_mp_mainnet,
        wallets=wallets,
        rotate_after_block=rotate_after_block,
        schain_log_level=schain_log_level,
        schain_log_level_config=schain_log_level_config
    ).to_config()
    schain_info = generate_schain_info(schain, schain_nodes)

    if base_config:
        update_base_config(base_config, schain, schain_nodes)
    else:
        base_config = {}
    return generate_schain_config(
        base_config=base_config,
        node_info=node_info,
        schain_info=schain_info
    )



import skale.contracts.data.nodes_data as nodes_data
from skale.dataclasses import CurrentNodeInfo
from skale.schain_config.generator import get_nodes_for_schain
from skale.schain_config.generator import (generate_schain_info, generate_schain_config,
                                           generate_skale_schain_config)
from skale.utils.helper import ip_from_bytes
from tests.constants import (DEFAULT_NODE_NAME, ZERO_ADDRESS, DEFAULT_SCHAIN_NAME, TEST_URL,
                             MIN_NODES_IN_SCHAIN)


TEST_NODE_IP_BYTES = b'\x8aD\xf6V'
TEST_NODE_IP = '10.10.10.10'
NODE_INFO_LEN = 15
SCHAIN_INFO_LEN = 4
TEST_ACCOUNTS_LEN = 2  # because we're creating everything from one account


TEST_BASE_CONFIG = {
    "params": {},
    "accounts": {}
}


def test_generate_node_info():
    node_info = CurrentNodeInfo(
        node_name=DEFAULT_NODE_NAME,
        node_id=0,
        base_port=10011,
        bind_ip=ip_from_bytes(TEST_NODE_IP_BYTES),
        ima_mainnet=TEST_URL,
        ima_mp_schain=ZERO_ADDRESS,
        ima_mp_mainnet=ZERO_ADDRESS,
        wallets={},
        rotate_after_block=128,
        schain_log_level_config='info'
    ).to_config()

    assert isinstance(node_info['bindIP'], str)
    assert isinstance(node_info['nodeID'], int)
    assert isinstance(node_info['nodeName'], str)
    assert isinstance(node_info['basePort'], int)
    assert isinstance(node_info['httpRpcPort'], int)
    assert isinstance(node_info['httpsRpcPort'], int)
    assert isinstance(node_info['wsRpcPort'], int)
    assert isinstance(node_info['wssRpcPort'], int)

    assert isinstance(node_info['imaMainNet'], str)
    assert isinstance(node_info['wallets'], dict)
    assert node_info['imaMessageProxySChain'] == ZERO_ADDRESS
    assert node_info['imaMessageProxyMainNet'] == ZERO_ADDRESS
    assert node_info['rotateAfterBlock'] == 128

    assert node_info['logLevel'] == 'trace'
    assert node_info['logLevelConfig'] == 'info'

    assert len(node_info) == NODE_INFO_LEN


def test_generate_schain_info():
    schain = {
        'name': DEFAULT_SCHAIN_NAME,
        'owner': ZERO_ADDRESS,
    }
    nodes = []
    schain_info = generate_schain_info(schain, nodes)
    assert isinstance(schain_info['schainID'], int)
    assert isinstance(schain_info['schainName'], str)
    assert isinstance(schain_info['schainOwner'], str)
    assert isinstance(schain_info['nodes'], list)
    assert len(schain_info) == SCHAIN_INFO_LEN


def test_generate_schain_config():
    config = generate_schain_config(TEST_BASE_CONFIG, {}, {})
    assert isinstance(config['accounts'], dict)
    assert isinstance(config['params'], dict)
    assert isinstance(config['skaleConfig']['nodeInfo'], dict)
    assert isinstance(config['skaleConfig']['sChain'], dict)


def test_get_nodes_for_schain(skale):
    schain_nodes = get_nodes_for_schain(skale, DEFAULT_SCHAIN_NAME)
    fields_with_id = nodes_data.FIELDS.copy()
    fields_with_id.append('id')

    assert len(schain_nodes) >= MIN_NODES_IN_SCHAIN
    assert list(schain_nodes[0].keys()) == fields_with_id


def test_generate_skale_schain_config(skale):
    node_index = skale.nodes_data.node_name_to_index(DEFAULT_NODE_NAME)
    config = generate_skale_schain_config(
        skale=skale,
        schain_name=DEFAULT_SCHAIN_NAME,
        node_id=node_index,
        base_config=TEST_BASE_CONFIG,
        ima_mainnet=TEST_URL,
        ima_mp_schain=ZERO_ADDRESS,
        ima_mp_mainnet=ZERO_ADDRESS,
        wallets={}
    )

    assert isinstance(config['params']['chainID'], str)
    assert len(config['accounts'].keys()) == TEST_ACCOUNTS_LEN

    custom_predeployed_address = '0xD2001000000000000000000000000000000000D2'
    assert isinstance(config['accounts'][custom_predeployed_address]['code'], str)
    assert len(config['accounts'][custom_predeployed_address]['storage']) == 2

    assert len(config['skaleConfig']['sChain']['nodes']) == 2
    assert isinstance(config['skaleConfig']['sChain']['schainOwner'], str)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

NODE_DEPOSIT = 100000000000000000000

GAS = {
    'add_validator': 4500000,
    'create_node': 8000000,
    'create_schain': 7500000,
    'get_bounty': 8000000,
    'send_verdict': 200000,
    'send_verdicts': 8000000,
    'set_check_time': 200000,
    'set_latency': 200000,
    'token_transfer': 600000,
    'add_authorized': 100000,
    'set_periods': 200000,
    'delete_node': 6000000,
    'delete_node_by_root': 6000000,
    'delete_schain': 8000000,
    'dkg_broadcast': 8000000,
    'dkg_response': 8000000,
    'dkg_alright': 1000000,
    'dkg_complaint': 1000000,

    'register_validator': 3000000,
    'delegate': 3000000,
    'accept_pending_delegation': 1000000,
    'enable_validator': 1000000,
    'disable_validator': 1000000,
    'link_node_address': 1000000,
    'unlink_node_address': 1000000,
    'cancel_pending_delegation': 1000000,
    'withdraw_bounty': 3000000,
    'withdraw_fee': 3000000,

    'skip_transition_delay': 1000000,
    'set_msr': 200000
}

OP_TYPES = {'create_node': 0x1, 'create_schain': 0x10}

LONG_LINE = '=' * 100

SCHAIN_TYPES = {
    'tiny': 1,
    'small': 2,
    'medium': 3,
    'test': 4,
}



# flake8: noqa: E402

from skale.wallets.common import BaseWallet
from skale.wallets.ledger_wallet import LedgerWallet
from skale.wallets.web3_wallet import Web3Wallet
from skale.wallets.rpc_wallet import RPCWallet
from skale.wallets.sgx_wallet import SgxWallet



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.contracts import BaseContract, transaction_method
from skale.utils.constants import GAS


class DKG(BaseContract):
    def gas_price(self):
        return self.skale.gas_price * 5 // 4

    @transaction_method(GAS['dkg_broadcast'])
    def broadcast(self, group_index, node_index,
                  verification_vector, secret_key_contribution):
        return self.contract.functions.broadcast(group_index, node_index,
                                                 verification_vector,
                                                 secret_key_contribution)

    @transaction_method(GAS['dkg_response'])
    def response(self, group_index, from_node_index,
                 secret_number, multiplied_share):
        return self.contract.functions.response(group_index, from_node_index,
                                                secret_number,
                                                multiplied_share)

    @transaction_method(GAS['dkg_alright'])
    def alright(self, group_index, from_node_index):
        return self.contract.functions.alright(group_index, from_node_index)

    @transaction_method(GAS['dkg_complaint'])
    def complaint(self, group_index, from_node_index, to_node_index):
        return self.contract.functions.complaint(group_index, from_node_index,
                                                 to_node_index)
